import nibabel as nb
import numpy as np


def compile_experiments(conditions, tasks):
    """
    Process conditions to compile a list of experiments and corresponding masks.

    Parameters
    ----------
    conditions : list of str
        Conditions for experiment selection:
        - `+tag`: Include experiments that have tag. Logical AND.
        - `-tag`: Exclude experiments that have tag. Logical NOT.
        - `?`: Combine included experiments (Logical OR).
        - `$file`: Load mask from file.

    tasks : pandas.DataFrame
        DataFrame with 'Name' and 'ExpIndex' columns for experiment lookup.

    Returns
    -------
    exp_to_use : list
        List of experiment indices to use.

    masks : list of numpy.ndarray
        List of masks from files.

    mask_names : list of str
        List of mask file names without extensions.
    """
    exp_to_use = []
    not_to_use = []
    masks = []
    mask_names = []

    for condition in conditions:
        operation = condition[0]
        argument = condition[1:]

        if operation == "+":
            # Logical AND
            tag = argument.lower()
            matches = tasks[tasks["Name"].str.lower() == tag]["ExpIndex"].values
            if not matches:
                raise ValueError(f"No experiments found for tag: {tag}")
            exp_to_use.append(matches[0])

        elif operation == "-":
            # Logical NOT
            tag = argument.lower()
            matches = tasks[tasks["Name"].str.lower() == tag]["ExpIndex"].values
            if not matches:
                raise ValueError(f"No experiments found for tag: {tag}")
            not_to_use.append(matches[0])

        elif operation == "?":
            # Logical OR: Combine all included experiments
            flat_list = [idx for sublist in exp_to_use for idx in sublist]
            exp_to_use = [list(set(flat_list))]

        elif operation == "$":
            # Load mask from file
            try:
                mask = nb.load(argument).get_fdata()
            except FileNotFoundError:
                raise FileNotFoundError(f"Mask file not found: {argument}")
            except Exception as e:
                raise ValueError(f"Error loading mask file {argument}: {e}")

            # Check binary or multi-class mask
            if np.unique(mask).size == 2:
                masks.append(mask.astype(bool))
            else:
                masks.append(mask.astype(int))

            mask_names.append(argument.rsplit(".", 1)[0])

    # Compute final set of experiments to use
    if exp_to_use:
        use_sets = map(set, exp_to_use)
        exp_to_use = list(set.intersection(*use_sets))
        if len(exp_to_use) == 0:
            raise ValueError("Bad tag combination. No experiments found.")
    else:
        raise ValueError("Bad tag combination. No experiments found.")

    # Remove excluded experiments
    if not_to_use:
        not_use_sets = map(set, not_to_use)
        not_to_use = list(set.union(*not_use_sets))
        exp_to_use = list(set(exp_to_use) - set(not_to_use))

    return exp_to_use, masks, mask_names



import logging
import tkinter

import customtkinter


class TextWidgetHandler(logging.Handler):
    def __init__(self, text_widget):
        super().__init__()
        self.text_widget = text_widget

    def emit(self, record):
        """
        Emit a log record to the text widget.
        """
        log_entry = self.format(record)
        self.text_widget.update_log(log_entry)


class OutputLogFrame(customtkinter.CTkFrame):
    def __init__(self, master, name):
        """
        Creates a 2x1 frame with a text box for outputting log messages.
        """
        super().__init__(master)
        self.name = name

        # configure grid layout (1x1)
        self.rowconfigure(0, weight=0)
        self.rowconfigure(1, weight=1)  # log row will resize
        self.columnconfigure(0, weight=1)

        # Frame title
        self.lbl_title = customtkinter.CTkLabel(
            master=self,
            text=f"{self.name}",
            justify=tkinter.LEFT,
        )
        self.lbl_title.grid(row=0, column=0, sticky="wns", padx=15, pady=15)

        # Text Box
        self.txt_log = tkinter.Text(
            master=self,
            wrap=tkinter.WORD,
            bg="#343638",
            fg="#ffffff",
            padx=20,
            pady=5,
            spacing1=4,  # spacing before a line
            spacing3=4,  # spacing after a line / wrapped line
            cursor="arrow",
            font=("Courier", 14),
        )
        self.txt_log.grid(row=1, column=0, padx=(15, 0), pady=(0, 15), sticky="nsew")
        self.txt_log.configure(state=tkinter.DISABLED)

        # Scrollbar
        self.scrollbar = customtkinter.CTkScrollbar(
            master=self, command=self.txt_log.yview
        )
        self.scrollbar.grid(row=1, column=1, padx=(0, 15), pady=(0, 15), sticky="ns")

        # Connect textbox scroll event to scrollbar
        self.txt_log.configure(yscrollcommand=self.scrollbar.set)

        self.controller = None

    def set_controller(self, controller):
        self.controller = controller

    def update_log(self, msg, overwrite=False):
        """
        Called from controller. Updates the log with given message. If overwrite is True,
        the last line will be cleared before the message is added.
        """
        self.txt_log.configure(state=tkinter.NORMAL)
        if overwrite:
            self.txt_log.delete("end-1c linestart", "end")
        self.txt_log.insert(tkinter.END, "\n" + msg)
        self.txt_log.configure(state=tkinter.DISABLED)
        self.txt_log.see(tkinter.END)

    def clear_log(self):
        """
        Called from controller. Clears the log.
        """
        self.txt_log.configure(state=tkinter.NORMAL)
        self.txt_log.delete(1.0, tkinter.END)
        self.txt_log.configure(state=tkinter.DISABLED)
        self.txt_log.see(tkinter.END)

    def set_logger(self, logger):
        """
        Attaches the logger to this frame's text widget.
        """
        self.logger = logger
        handler = TextWidgetHandler(self)
        handler.setFormatter(logging.Formatter("%(levelname)s - %(message)s"))
        if len(logger.handlers) < 3:
            self.logger.addHandler(handler)
            self.logger.setLevel(logging.DEBUG)  # Set the desired log level



import logging

import matplotlib.pyplot as plt
import numpy as np

from jale.core.utils.compute import compute_ale, compute_ma
from jale.core.utils.folder_setup import folder_setup
from jale.core.utils.template import GM_SAMPLE_SPACE

logger = logging.getLogger("ale_logger")


def roi_ale(project_path, exp_df, meta_name, mask, mask_name, monte_carlo_iterations):
    # Stack the modeled activation (MA) values and compute ALE values
    folder_setup(project_path, "ROI")
    ma = np.load(project_path / f"MainEffect/{meta_name}_ma.npz")["arr_0"]
    ale = compute_ale(ma)

    # Calculate benchmark values within the mask
    mask_sum = np.sum(ale[mask])
    mask_max = np.max(ale[mask])

    # Initialize null distribution array for masked ALE values
    mask_size = np.sum(mask > 0)
    null_ale_mask = np.zeros((monte_carlo_iterations, mask_size))
    num_peaks = exp_df.NumberOfFoci.values

    # Simulate null distribution with repeated random sampling
    for repeat in range(monte_carlo_iterations):
        if repeat > 0 and repeat % 1000 == 0:
            logger.info(f"Simulated {repeat} iterations of ROI null distribution")

        # Generate random peaks and compute null ALE values
        null_peaks = np.array(
            [
                GM_SAMPLE_SPACE[
                    :, np.random.randint(0, GM_SAMPLE_SPACE.shape[1], num_peak)
                ].T
                for num_peak in num_peaks
            ],
            dtype=object,
        )

        null_ma = compute_ma(null_peaks, exp_df.Kernels)
        null_ale = compute_ale(null_ma)
        null_ale_mask[repeat] = null_ale[mask]

    # Aggregate null distribution statistics
    null_sum = np.sum(null_ale_mask, axis=1)
    null_max = np.max(null_ale_mask, axis=1)

    # Plot the results
    plot_roi_ale(
        project_path,
        meta_name,
        mask_name,
        mask_sum,
        mask_max,
        null_sum,
        null_max,
        monte_carlo_iterations,
    )


def plot_roi_ale(
    project_path,
    meta_name,
    mask_name,
    bench_sum,
    bench_max,
    null_sum,
    null_max,
    monte_carlo_iterations,
):
    fig, ax = plt.subplots(1, 2, figsize=(15, 10))
    fig.patch.set_facecolor("skyblue")

    def plot_histogram(axis, data, bench_value, title, xlabel):
        weights = np.ones_like(data) / len(data)
        n, _, patches = axis.hist(
            data, bins=np.linspace(0, np.ceil(np.max(data)), 50), weights=weights
        )

        # Plot benchmark line and annotate p-value
        axis.vlines(bench_value, 0, np.max(n) + np.max(n) / 8, colors="r")
        p_value = (data > bench_value).sum() / monte_carlo_iterations
        axis.annotate(
            f"Observed value\np-value = {p_value:.3f}",
            xy=(bench_value, np.max(n) + np.max(n) / 8),
            ha="center",
        )

        # Plot significance lines and percentiles
        percentiles = [
            (95, "darkgreen", "p < 0.05"),
            (99, "green", "p < 0.01"),
            (99.9, "lime", "p < 0.001"),
        ]

        for perc, color, label in percentiles:
            axis.vlines(
                np.percentile(data, perc), 0, np.max(n), colors=color, label=label
            )

        axis.set_xlabel(xlabel)
        axis.set_ylabel(f"Percentage of {monte_carlo_iterations} realizations")
        axis.title.set_text(title)
        axis.legend(loc="upper right")

    # Plot ALE integral histogram
    plot_histogram(
        ax[0],
        null_sum,
        bench_sum,
        f"{mask_name} - ALE integral",
        "ALE integral in ROI mask",
    )

    # Plot max ALE histogram
    ax[1].yaxis.tick_right()
    plot_histogram(
        ax[1],
        null_max,
        bench_max,
        f"{mask_name} - max ALE",
        "Max ALE in ROI mask",
    )

    fig.tight_layout()

    save_path = project_path / f"Results/MainEffect/ROI/{meta_name}_{mask_name}.png"
    fig.savefig(save_path)
    plt.close()



import sys
import tkinter

import customtkinter

# CustomText widget for displaying the output


class CustomText(tkinter.Text):
    def __init__(self, master, **kwargs):
        super().__init__(master, **kwargs)
        self.configure(state=tkinter.DISABLED)
        # custom_font = tkinter.font.Font(family="Helvetica", size=12)
        # self.configure(font=custom_font)

    def write(self, message):
        self.configure(state=tkinter.NORMAL)
        self.insert(tkinter.END, message)
        self.configure(state=tkinter.DISABLED)
        self.see(tkinter.END)

    def flush(self):
        pass


# Redirect standard output to the CustomText widget


class RedirectedStdout:
    def __init__(self, text_widget):
        self.text_widget = text_widget

    def write(self, message):
        self.text_widget.write(message)

    def flush(self):
        pass


# Frame to hold the CustomText widget


class OutputFrame(customtkinter.CTkFrame):
    def __init__(self, master, **kwargs):
        super().__init__(master, **kwargs)
        self.grid_rowconfigure(0, weight=1)
        self.grid_columnconfigure(0, weight=1)

        self.text_widget = CustomText(self)
        self.text_widget.grid(row=0, column=0, sticky="nsew")

        self.scrollbar = customtkinter.CTkScrollbar(
            self, command=self.text_widget.yview
        )
        self.scrollbar.grid(row=0, column=1, sticky="ns")
        self.text_widget["yscrollcommand"] = self.scrollbar.set

        self.redirect_stdout()

    def redirect_stdout(self):
        sys.stdout = RedirectedStdout(self.text_widget)



import tkinter
import tkinter.messagebox as messagebox

import customtkinter


class ParameterWindow(customtkinter.CTkToplevel):
    def __init__(self, master, controller):
        super().__init__(master)
        self.controller = controller
        self.title("ALE Parameters")
        self.grid_columnconfigure(0, weight=1)

        self.title_label = customtkinter.CTkLabel(
            self,
            text="Parameter Settings",
            font=customtkinter.CTkFont(size=20, weight="bold"),
        )
        self.title_label.grid(row=0, column=0, padx=20, pady=10)

        # Cutoff prediction checkbox
        self.cutoff_prediction_var = tkinter.BooleanVar(value=True)
        self.cutoff_prediction_checkbox = customtkinter.CTkCheckBox(
            self, text="Cutoff Prediction", variable=self.cutoff_prediction_var
        )
        self.cutoff_prediction_checkbox.grid(
            row=1, column=0, padx=20, pady=10, sticky="w"
        )

        # TFCE enabled checkbox
        self.tfce_enabled_var = tkinter.BooleanVar(value=True)
        self.tfce_enabled_checkbox = customtkinter.CTkCheckBox(
            self, text="TFCE Enabled", variable=self.tfce_enabled_var
        )
        self.tfce_enabled_checkbox.grid(row=2, column=0, padx=20, pady=10, sticky="w")

        # Monte-carlo iterations textbox
        self.monte_carlo_iterations_label = customtkinter.CTkLabel(
            self, text="Monte-Carlo Iterations"
        )
        self.monte_carlo_iterations_label.grid(
            row=3, column=0, padx=20, pady=(10, 0), sticky="w"
        )
        self.monte_carlo_iterations_textbox = customtkinter.CTkEntry(self)
        self.monte_carlo_iterations_textbox.insert(0, "5000")
        self.monte_carlo_iterations_textbox.grid(
            row=4, column=0, padx=20, pady=(0, 10), sticky="ew"
        )

        # cFWE cluster-forming threshold textbox (0.00001, 0.0001, 0.001, 0.005, 0.01)
        self.cfwe_threshold_label = customtkinter.CTkLabel(
            self, text="cFWE Cluster-Forming Threshold"
        )
        self.cfwe_threshold_label.grid(
            row=5, column=0, padx=20, pady=(10, 0), sticky="w"
        )
        self.cfwe_threshold_textbox = customtkinter.CTkEntry(self)
        self.cfwe_threshold_textbox.insert(0, "0.001")
        self.cfwe_threshold_textbox.grid(
            row=6, column=0, padx=20, pady=(0, 10), sticky="ew"
        )

        # Subsampling N textbox (500,1000,2500,5000,10000)
        self.subsampling_n_label = customtkinter.CTkLabel(
            self, text="IPA Subsampling N"
        )
        self.subsampling_n_label.grid(
            row=7, column=0, padx=20, pady=(10, 0), sticky="w"
        )
        self.subsampling_n_textbox = customtkinter.CTkEntry(self)
        self.subsampling_n_textbox.insert(0, "2500")
        self.subsampling_n_textbox.grid(
            row=8, column=0, padx=20, pady=(0, 10), sticky="ew"
        )

        # Balanced contrast monte-carlo iterations textbox (500,1000,2000,3000,5000)
        self.balanced_contrast_label = customtkinter.CTkLabel(
            self, text="Balanced Contrast Monte-Carlo Iterations"
        )
        self.balanced_contrast_label.grid(
            row=9, column=0, padx=20, pady=(10, 0), sticky="w"
        )
        self.balanced_contrast_textbox = customtkinter.CTkEntry(self)
        self.balanced_contrast_textbox.insert(0, "1000")
        self.balanced_contrast_textbox.grid(
            row=10, column=0, padx=20, pady=(0, 10), sticky="ew"
        )

        # Apply button
        self.apply_button = customtkinter.CTkButton(
            self, text="Apply", command=self.apply_parameters
        )
        self.apply_button.grid(row=11, column=0, padx=20, pady=10)

    def validate_monte_carlo_iterations(self):
        valid = True
        try:
            value = int(self.monte_carlo_iterations_textbox.get())
            if value < 100 or value > 100000:
                messagebox.showerror(
                    "Invalid Input",
                    "Monte Carlo iterations must be a whole number between 100 and 100,000.",
                )
                self.monte_carlo_iterations_textbox.delete(0, "end")
                self.monte_carlo_iterations_textbox.insert(0, "5000")
                valid = False
        except ValueError:
            messagebox.showerror(
                "Invalid Input",
                "Monte Carlo iterations must be a whole number between 100 and 100,000.",
            )
            self.monte_carlo_iterations_textbox.delete(0, "end")
            self.monte_carlo_iterations_textbox.insert(0, "5000")
            valid = False
        return valid

    def validate_cluster_forming_threshold(self):
        valid = True
        try:
            value = float(self.cfwe_threshold_textbox.get())
            if value < 0.00001 or value > 0.05:
                messagebox.showerror(
                    "Invalid Input",
                    "cFWE cluster forming threshold must be between 0.00001 and 0.05.",
                )
                self.cfwe_threshold_textbox.delete(0, "end")
                self.cfwe_threshold_textbox.insert(0, "0.001")
                valid = False
        except ValueError:
            messagebox.showerror(
                "Invalid Input",
                "cFWE cluster forming threshold must be between 0.00001 and 0.05.",
            )
            self.cfwe_threshold_textbox.delete(0, "end")
            self.cfwe_threshold_textbox.insert(0, "0.001")
            valid = False
        return valid

    def validate_subsampling_n(self):
        try:
            value = int(self.subsampling_n_textbox.get())
            if value < 100 or value > 25000:
                messagebox.showerror(
                    "Invalid Input",
                    "IPA subsampling N must be a whole number between 100 and 25000.",
                )
                self.subsampling_n_textbox.delete(0, "end")
                self.subsampling_n_textbox.insert(0, "2500")
                return False
        except ValueError:
            messagebox.showerror(
                "Invalid Input",
                "IPA subsampling N must be a whole number between 100 and 25000.",
            )
            self.subsampling_n_textbox.delete(0, "end")
            self.subsampling_n_textbox.insert(0, "2500")
            return False

    def validate_balanced_contrast_iterations(self):
        valid = True
        try:
            value = int(self.balanced_contrast_textbox.get())
            if value < 100 or value > 10000:
                messagebox.showerror(
                    "Invalid Input",
                    "Balanced contrast monte-carlo iterations must be a whole number between 100 and 10000.",
                )
                self.balanced_contrast_textbox.delete(0, "end")
                self.balanced_contrast_textbox.insert(0, "1000")
                valid = False
        except ValueError:
            messagebox.showerror(
                "Invalid Input",
                "Balanced contrast monte-carlo iterations must be a whole number between 100 and 10000.",
            )
            self.balanced_contrast_textbox.delete(0, "end")
            self.balanced_contrast_textbox.insert(0, "1000")
            valid = False
        return valid

    def apply_parameters(self):
        val_mc = self.validate_monte_carlo_iterations()
        val_cft = self.validate_cluster_forming_threshold()
        val_subn = self.validate_subsampling_n()
        val_bci = self.validate_balanced_contrast_iterations()

        if any([val_mc, val_cft, val_subn, val_bci]) is False:
            return

        parameters = {
            "cutoff_prediction": self.cutoff_prediction_var.get(),
            "tfce_enabled": self.tfce_enabled_var.get(),
            "monte_carlo_iterations": int(self.monte_carlo_iterations_textbox.get()),
            "cfwe_threshold": float(self.cfwe_threshold_textbox.get()),
            "subsampling_n": int(self.subsampling_n_textbox.get()),
            "balanced_contrast_iterations": int(self.balanced_contrast_textbox.get()),
        }
        self.controller.get_ale_parameters(parameters)

        self.apply_button.configure(fg_color="green")
        self.apply_button.configure(text="Applied")
        print("Changed ALE Parameters")
        self.after(
            2000, lambda: self.apply_button.configure(fg_color="#20548c", text="Apply")
        )


class ParameterWarningWindow(customtkinter.CTkToplevel):
    def __init__(self, master, controller):
        super().__init__(master)
        self.controller = controller
        self.title("WARNING!")
        self.warning_label = customtkinter.CTkLabel(
            self,
            text="Default parameter values are set automatically.\nChanging parameters only advised for experts.",
            font=customtkinter.CTkFont(size=16),
        )
        self.warning_label.grid(row=0, column=0, columnspan=2, padx=20, pady=10)

        self.understand_button = customtkinter.CTkButton(
            self, text="I understand.", command=self.understand_button_event
        )
        self.understand_button.grid(row=1, column=0, padx=20, pady=10)

        self.close_button = customtkinter.CTkButton(
            self, text="Close", command=self.close_button_event
        )
        self.close_button.grid(row=1, column=1, padx=20, pady=10)

    def understand_button_event(self):
        self.controller.open_parameter_window()
        self.destroy()

    def close_button_event(self):
        self.destroy()



import operator
from functools import reduce

import numpy as np
from scipy import ndimage


def tfce_par(invol, h, dh, voxel_dims=[2, 2, 2], E=0.6, H=2):
    """
    Compute the Threshold-Free Cluster Enhancement (TFCE) update values for a given threshold.

    This function applies a threshold to an input volume, identifies clusters above the threshold,
    and calculates TFCE update values based on cluster sizes and specified parameters.

    Parameters
    ----------
    invol : numpy.ndarray
        Input 3D volume to be thresholded and analyzed.
    h : float
        Intensity threshold applied to `invol`.
    dh : float
        Step size for threshold increments in TFCE calculation.
    voxel_dims : list of int, optional
        Dimensions of each voxel in millimeters, by default [2, 2, 2].
    E : float, optional
        TFCE parameter that adjusts the emphasis on cluster size, by default 0.6.
    H : float, optional
        TFCE parameter that adjusts the emphasis on intensity height, by default 2.

    Returns
    -------
    tuple
        - numpy.ndarray : TFCE update values calculated for suprathreshold clusters.
        - numpy.ndarray : Mask of labeled suprathreshold areas.
    """
    # Threshold input volume based on intensity threshold `h`
    thresh = invol > h

    # Identify suprathreshold clusters and get the number of unique clusters
    labels, cluster_count = ndimage.label(thresh)  # type: ignore

    # Calculate sizes of each cluster, converting to voxel volume by multiplying dimensions
    sizes = np.bincount(labels.ravel()) * reduce(operator.mul, voxel_dims)

    # Create a mask of only the labeled (clustered) areas
    mask = labels > 0

    # Compute TFCE update values for each cluster using the specified E and H parameters
    update_vals = (h**H) * dh * (sizes[labels[mask]] ** E)

    return update_vals, mask



import math

import numpy as np

from jale.core.utils.template import PAD_SHAPE


def create_kernel_array(exp_df):
    """
    Create an array of 3D Gaussian kernels based on subject count for each experiment.

    This function calculates a smoothing kernel for each experiment in the dataset,
    taking into account both template and subject uncertainties. The kernel is stored
    in a 3D array representing the Gaussian spread for each experiment.

    Parameters
    ----------
    exp_df : pandas.DataFrame
        DataFrame containing experiment information with a 'Subjects' column.

    Returns
    -------
    numpy.ndarray
        4D array of Gaussian kernels, with shape (num_experiments, 31, 31, 31).
    """
    # Initialize an empty array for storing kernels, with fixed dimensions (31, 31, 31)
    kernel_array = np.empty((exp_df.shape[0], 31, 31, 31))

    # Calculate template uncertainty (constant across experiments)
    template_uncertainty = 5.7 / (2 * np.sqrt(2 / np.pi)) * np.sqrt(8 * np.log(2))

    # Generate a kernel for each experiment based on subject-specific uncertainty
    for i, n_subjects in enumerate(exp_df.Subjects.values):
        # Calculate subject uncertainty based on the number of subjects
        subj_uncertainty = (
            11.6 / (2 * np.sqrt(2 / np.pi)) * np.sqrt(8 * np.log(2))
        ) / np.sqrt(n_subjects)

        # Combine uncertainties to get smoothing value for the Gaussian kernel
        smoothing = np.sqrt(template_uncertainty**2 + subj_uncertainty**2)

        # Compute and store the 3D kernel for the current experiment
        kernel_array[i] = compute_3dkernel(smoothing, 31)

    return kernel_array


def compute_3dkernel(fwhm, dims):
    """
    Compute a 3D Gaussian kernel based on full width at half maximum (FWHM).

    This function generates a 3D Gaussian kernel by convolving three 1D Gaussians.
    The kernel is padded to fit the specified dimensions.

    Parameters
    ----------
    fwhm : float
        Full width at half maximum, used to calculate the Gaussian spread.
    dims : int
        Desired size for the output kernel (dims x dims x dims).

    Returns
    -------
    numpy.ndarray
        3D Gaussian kernel array with specified dimensions.
    """
    # Calculate the variance from FWHM
    s = (
        fwhm / 2 / math.sqrt(8 * math.log(2)) + np.finfo(float).eps
    ) ** 2  # fwhm -> sigma

    # Define kernel range in one dimension
    half_k_length = math.ceil(3.5 * math.sqrt(s))
    x = list(range(-half_k_length, half_k_length + 1))

    # Create a normalized 1D Gaussian kernel
    oned_kernel = np.exp(-0.5 * np.multiply(x, x) / s) / math.sqrt(2 * math.pi * s)
    oned_kernel /= np.sum(oned_kernel)

    # Create a 3D Gaussian by convolving 1D Gaussian along each axis
    gkern3d = (
        oned_kernel[:, None, None]
        * oned_kernel[None, :, None]
        * oned_kernel[None, None, :]
    )

    # Pad the 3D Gaussian to match desired dimensions
    pad_size = int((dims - len(x)) / 2)
    gkern3d = np.pad(
        gkern3d,
        ((pad_size, pad_size), (pad_size, pad_size), (pad_size, pad_size)),
        "constant",
        constant_values=0,
    )
    return gkern3d


def kernel_convolution(foci, kernel):
    """
    Apply a Gaussian kernel to foci coordinates within a 3D space.

    This function convolves a Gaussian kernel over a 3D array, centered at each
    specified focus point, using maximum value in overlapping areas.

    Parameters
    ----------
    foci : list of tuples
        List of coordinates where each focus point is applied in 3D space.
    kernel : numpy.ndarray
        3D Gaussian kernel array used for convolution.

    Returns
    -------
    numpy.ndarray
        3D array with kernel applied at each focus point, trimmed to the padded size.
    """
    # Determine kernel size and padding required for each axis
    kernel_size = kernel.shape[0]
    pad = kernel_size // 2

    # Initialize a 3D array with padding around the focus coordinates
    data = np.zeros(PAD_SHAPE)

    # Iterate over each focus point and apply the kernel centered at the focus
    for focus in foci:
        # Define the boundaries of the region where the kernel will be applied
        x_start, x_end = focus[0], focus[0] + kernel_size
        y_start, y_end = focus[1], focus[1] + kernel_size
        z_start, z_end = focus[2], focus[2] + kernel_size

        # Use maximum operation for overlapping kernel regions
        data[x_start:x_end, y_start:y_end, z_start:z_end] = np.maximum(
            data[x_start:x_end, y_start:y_end, z_start:z_end], kernel
        )

    # Trim padding and return the final 3D array
    return data[
        pad : data.shape[0] - pad, pad : data.shape[1] - pad, pad : data.shape[2] - pad
    ]



import logging
import pickle
from pathlib import Path

import nibabel as nb
import numpy as np
import pandas as pd
from joblib import Parallel, delayed

from jale.core.utils.compute import (
    compute_ale,
    compute_clusters,
    compute_hx,
    compute_hx_conv,
    compute_ma,
    compute_monte_carlo_null,
    compute_tfce,
    compute_z,
    illustrate_foci,
)
from jale.core.utils.cutoff_prediction import predict_cutoff
from jale.core.utils.folder_setup import folder_setup
from jale.core.utils.kernel import create_kernel_array
from jale.core.utils.plot_and_save import plot_and_save
from jale.core.utils.template import GM_PRIOR

logger = logging.getLogger("ale_logger")


def main_effect(
    project_path,
    exp_df,
    meta_name,
    tfce_enabled=True,
    cutoff_predict_enabled=True,
    bin_steps=0.0001,
    cluster_forming_threshold=0.001,
    monte_carlo_iterations=5000,
    nprocesses=2,
):
    """
    Compute and save the main effect map for a given meta-analysis.

    This function calculates the main effect ALE, going from a coordinate dataframe all the way to thresholded brain maps.
    Uses either monte-carlo simulation or machine learning prediction for multiple comparison correction (vFWE, cFWE or TFCE).

    Parameters
    ----------
    project_path : str or Path
        Path to the project directory containing the "Results" folder.
    exp_df : pandas.DataFrame
        DataFrame containing experiment data, including coordinates and number of foci.
    meta_name : str
        Name of the meta-analysis, used for naming saved files.
    tfce_enabled : bool, optional
        Whether to compute TFCE-corrected maps, by default True.
    cutoff_predict_enabled : bool, optional
        If True, predicts statistical thresholds using ML models, by default True.
    bin_steps : float, optional
        Step size for defining histogram bins, by default 0.0001.
    cluster_forming_threshold : float, optional
        Threshold for forming clusters in ALE, by default 0.001.
    monte_carlo_iterations : int, optional
        Number of Monte Carlo iterations for null distribution simulation, by default 5000.
    target_n : int, optional
        Target number of subsamples for probabilistic ALE, by default None (uses full sample).
    sample_n : int, optional
        Number of subsamples to generate if `target_n` is specified, by default 2500.
    nprocesses : int, optional
        Number of parallel processes for computations, by default 2.

    Returns
    -------
    None
        The function performs computations and saves the results as NIfTI files in the
        specified `project_path` directory.
    """

    logger.info(
        f"{meta_name} : {exp_df.shape[0]} experiments; {exp_df.Subjects.sum()}"
        f" total subjects - average of {exp_df.Subjects.mean():.2f} subjects per experiment"
    )

    # set main_effect results folder as path
    folder_setup(project_path, "MainEffect")
    project_path = (Path(project_path) / "Results/MainEffect/").resolve()

    # calculate smoothing kernels for each experiment
    kernels = create_kernel_array(exp_df)
    np.save(project_path / f"{meta_name}_kernels", kernels)

    # calculate maximum possible ale value to set boundaries for histogram bins
    max_ma = np.prod([1 - np.max(kernel) for kernel in kernels])

    # define bins for histogram
    bin_edges = np.arange(0.00005, 1 - max_ma + 0.001, bin_steps)
    bin_centers = np.arange(0, 1 - max_ma + 0.001, bin_steps)
    step = int(1 / bin_steps)

    # Save included experiments for provenance tracking
    print_df = pd.DataFrame(
        {
            "Experiment": exp_df.Articles.values,
            "Number of Foci": exp_df.NumberOfFoci.values,
        }
    )
    print_df.to_csv(
        project_path / f"{meta_name}_included_experiments.csv", index=False, sep="\t"
    )

    # MA calculation
    ma = compute_ma(exp_df.Coordinates.values, kernels)
    np.savez_compressed(project_path / f"{meta_name}_ma", ma)

    # Foci illustration
    if not Path(project_path / f"Volumes/{meta_name}_foci.nii").exists():
        logger.info(f"{meta_name} - illustrate Foci")
        # take all peaks of included studies and save them in a Nifti
        foci_arr = illustrate_foci(exp_df.Coordinates.values)
        plot_and_save(project_path, f"{meta_name}_foci", foci_arr)

    # ALE calculation
    if Path(project_path / f"NullDistributions/{meta_name}.pickle").exists():
        logger.info(f"{meta_name} - loading ALE")
        logger.info(f"{meta_name} - loading null PDF")
        ale = nb.loadsave.load(
            project_path / f"Volumes/{meta_name}_ale.nii"
        ).get_fdata()  # type: ignore
        with open(project_path / f"NullDistributions/{meta_name}.pickle", "rb") as f:
            hx_conv, _ = pickle.load(f)

    else:
        logger.info(f"{meta_name} - computing ALE and null PDF")
        ale = compute_ale(ma)
        plot_and_save(project_path, f"{meta_name}_ale", ale)

        # Calculate histogram and use it to estimate a null probability density function
        hx = compute_hx(ma, bin_edges)
        hx_conv = compute_hx_conv(hx, bin_centers, step)

        pickle_object = (hx_conv, hx)
        with open(
            project_path / f"NullDistributions/{meta_name}_histogram.pickle",
            "wb",
        ) as f:
            pickle.dump(pickle_object, f)

    # z- and tfce-map calculation
    if Path(project_path / f"Volumes/{meta_name}_z.nii").exists():
        logger.info(f"{meta_name} - loading z-values & TFCE")
        z = nb.loadsave.load(project_path / f"Volumes/{meta_name}_z.nii").get_fdata()  # type: ignore

    else:
        logger.info(f"{meta_name} - computing p-values & TFCE")
        z = compute_z(ale, hx_conv, step)
        plot_and_save(project_path, f"{meta_name}_z", z)
    if tfce_enabled is True:
        if Path(project_path / f"Volumes/{meta_name}_tfce_uncorrected.nii").exists():
            tfce = nb.loadsave.load(
                project_path / f"Volumes/{meta_name}_tfce_uncorrected.nii"
            ).get_fdata()  # type: ignore
        else:
            tfce = compute_tfce(z)
            plot_and_save(project_path, f"{meta_name}_tfce_uncorrected", tfce)

    # monte-carlo simulation for multiple comparison corrected thresholds
    if cutoff_predict_enabled:
        # using ml models to predict thresholds
        vfwe_treshold, cfwe_threshold, tfce_threshold = predict_cutoff(exp_df=exp_df)
    else:
        if Path(
            project_path / f"NullDistributions/{meta_name}_montecarlo.pickle"
        ).exists():
            logger.info(f"{meta_name} - loading null")
            with open(
                project_path / f"NullDistributions/{meta_name}_montecarlo.pickle",
                "rb",
            ) as f:
                vfwe_null, cfwe_null, tfce_null = pickle.load(f)
        else:
            logger.info(f"{meta_name} - simulating null")
            vfwe_null, cfwe_null, tfce_null = zip(
                *Parallel(n_jobs=nprocesses, verbose=2)(
                    delayed(compute_monte_carlo_null)(
                        num_foci=exp_df.NumberOfFoci,
                        kernels=kernels,
                        bin_edges=bin_edges,
                        bin_centers=bin_centers,
                        step=step,
                        cluster_forming_threshold=cluster_forming_threshold,
                        tfce_enabled=tfce_enabled,
                    )
                    for i in range(monte_carlo_iterations)
                )
            )

            simulation_pickle = (vfwe_null, cfwe_null, tfce_null)
            with open(
                project_path / f"NullDistributions/{meta_name}_montecarlo.pickle",
                "wb",
            ) as f:
                pickle.dump(simulation_pickle, f)

        vfwe_treshold = np.percentile(vfwe_null, 95)
        cfwe_threshold = np.percentile(cfwe_null, 95)
        tfce_threshold = np.percentile(tfce_null, 95)

    # Tresholding maps with vFWE, cFWE, TFCE thresholds
    if not Path(project_path / f"Volumes/{meta_name}_vFWE.nii").exists():
        logger.info(f"{meta_name} - inference and printing")
        # voxel wise family wise error correction
        ale[GM_PRIOR == 0] = 0
        vfwe_map = ale * (ale > vfwe_treshold)
        plot_and_save(project_path, f"{meta_name}_vFWE", vfwe_map)
        if np.max(ale) > vfwe_treshold:
            logger.info("vFWE: significant effect found.")
        else:
            logger.info("vFWE: no significant effect found.")

        # cluster wise family wise error correction
        z[GM_PRIOR == 0] = 0
        cfwe_map, max_clust = compute_clusters(
            z, cluster_forming_threshold, cfwe_threshold
        )
        plot_and_save(project_path, f"{meta_name}_cFWE", cfwe_map)
        if max_clust > cfwe_threshold:
            logger.info("cFWE: significant effect found.")
        else:
            logger.info("cFWE: no significant effect found.")

        # tfce error correction
        if tfce_enabled:
            tfce[GM_PRIOR == 0] = 0
            tfce_map = tfce * (tfce > tfce_threshold)
            plot_and_save(project_path, f"{meta_name}_tfce", tfce_map)
            if np.max(tfce) > tfce_threshold:
                logger.info("TFCE: significant effect found.")
            else:
                logger.info("TFCE: no significant effect found.")

    else:
        pass
        logger.info(f"{meta_name} - done!")



import logging
from multiprocessing import Process
from threading import Thread

import customtkinter

from jale.ale import (
    load_config,
    load_dataframes,
    run_balanced_contrast,
    run_contrast_analysis,
    run_ma_clustering,
    run_main_effect,
    run_probabilistic_ale,
    setup_logger,
    setup_project_folder,
)


class Controller:
    def __init__(
        self, sidebar_frame, analysis_table_frame, dataset_table_frame, output_log_frame
    ):
        # Frames
        self.sidebar_frame = sidebar_frame
        self.analysis_table_frame = analysis_table_frame
        self.dataset_table_frame = dataset_table_frame
        self.output_log_frame = output_log_frame

        # ALE objects
        self.analysis_df = None
        self.dataset_df = None
        self.task_df = None
        self.params = None
        self.clustering_params = None
        self.config = None
        self.project_path = None

    # Sidebar Buttons
    def config_to_gui(self):
        filename = customtkinter.filedialog.askopenfilename()
        if filename:
            self.config = load_config(filename)
            self.project_path = setup_project_folder(self.config)

            logger = setup_logger(self.project_path)
            self.output_log_frame.set_logger(logger)
            logger.info("Logger initialized and project setup complete.")
            self.params = self.config.get("parameters", {})
            self.clustering_params = self.config.get("clustering_params", {})
            self.dataset_df, self.task_df, self.analysis_df = load_dataframes(
                self.project_path, self.config
            )
            self.dataset_table_frame.fill_dataset_table(self.dataset_df)
            analysis_df_formatted = (
                self.analysis_table_frame.format_imported_analysis_file(
                    self.analysis_df
                )
            )
            self.analysis_table_frame.fill_analysis_table(analysis_df_formatted)

    def run_analysis(self):
        logger = logging.getLogger("ale_logger")

        def process_analyses():
            """
            Run analyses sequentially in threads.
            """
            for row_idx in range(self.analysis_df.shape[0]):
                # Skip empty rows
                if not isinstance(self.analysis_df.iloc[row_idx, 0], str):
                    continue

                thread = None

                if self.analysis_df.iloc[row_idx, 0] == "M":
                    thread = Thread(
                        target=run_main_effect,
                        args=(
                            self.analysis_df,
                            row_idx,
                            self.project_path,
                            self.params,
                            self.dataset_df,
                            self.task_df,
                        ),
                    )

                elif self.analysis_df.iloc[row_idx, 0][0] == "P":
                    thread = Thread(
                        target=run_probabilistic_ale,
                        args=(
                            self.analysis_df,
                            row_idx,
                            self.project_path,
                            self.params,
                            self.dataset_df,
                            self.task_df,
                        ),
                    )

                elif self.analysis_df.iloc[row_idx, 0] == "C":
                    thread = Thread(
                        target=run_contrast_analysis,
                        args=(
                            self.analysis_df,
                            row_idx,
                            self.project_path,
                            self.params,
                            self.dataset_df,
                            self.task_df,
                        ),
                    )

                elif self.analysis_df.iloc[row_idx, 0][0] == "B":
                    thread = Thread(
                        target=run_balanced_contrast,
                        args=(
                            self.analysis_df,
                            row_idx,
                            self.project_path,
                            self.params,
                            self.dataset_df,
                            self.task_df,
                        ),
                    )

                elif self.analysis_df.iloc[row_idx, 0] == "Cluster":
                    thread = Thread(
                        target=run_ma_clustering,
                        args=(
                            self.analysis_df,
                            row_idx,
                            self.project_path,
                            self.clustering_params,
                            self.dataset_df,
                            self.task_df,
                        ),
                    )

                if thread:
                    thread.start()
                    thread.join()  # Wait for the current thread to finish

            logger.info("Analysis completed.")

        # Run analyses in a background thread to keep the GUI responsive
        background_thread = Thread(target=process_analyses)
        background_thread.start()

    def stop_analysis(self):
        return



import tkinter as tk
from tkinter import ttk

import customtkinter
import pandas as pd


class AnalysisTableFrame(customtkinter.CTkFrame):
    def __init__(self, master):
        super().__init__(master)
        self.analysis_type_abbreviation_dict = {
            "Main Effect": "M",
            "IPA": "P",
            "Standard Contrast": "C",
            "Balanced Contrast": "B",
        }

        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)

        self.lbl_title = customtkinter.CTkLabel(
            master=self,
            text="Analysis Info",
            justify=tk.LEFT,
            fg_color="#477AA2",
            corner_radius=5,
        )
        self.lbl_title.grid(row=0, column=0, sticky="ew", pady=(0, 1))

        self.tree_frame = customtkinter.CTkFrame(self)
        self.tree_frame.grid(row=1, column=0, sticky="nsew")

        self.tree_frame.grid_rowconfigure(0, weight=1)
        self.tree_frame.grid_columnconfigure(0, weight=1)

        self.tree_scroll = ttk.Scrollbar(self.tree_frame)
        self.tree_scroll.grid(row=0, column=1, sticky="ns")

        self.tree = ttk.Treeview(self.tree_frame, yscrollcommand=self.tree_scroll.set)
        self.tree.grid(row=0, column=0, sticky="nsew")

        self.tree_scroll.config(command=self.tree.yview)

        self.tree["columns"] = (
            "Analysis Type",
            "Analysis Name",
            "Group 1 Logic",
            "Group 2 Logic",
        )

        self.tree.column("#0", width=0, stretch=tk.NO)
        self.tree.column("Analysis Type", anchor=tk.CENTER, width=100, stretch=tk.NO)
        self.tree.column("Analysis Name", anchor=tk.CENTER, width=250, stretch=tk.NO)
        self.tree.column("Group 1 Logic", anchor=tk.CENTER, width=280, stretch=tk.NO)
        self.tree.column("Group 2 Logic", anchor=tk.CENTER, width=280, stretch=tk.NO)

        self.tree.heading("#0", text="", anchor=tk.CENTER)
        self.tree.heading("Analysis Type", text="Analysis Type", anchor=tk.CENTER)
        self.tree.heading("Analysis Name", text="Analysis Name", anchor=tk.CENTER)
        self.tree.heading("Group 1 Logic", text="Group 1 Logic", anchor=tk.CENTER)
        self.tree.heading("Group 2 Logic", text="Group 2 Logic", anchor=tk.CENTER)

    def set_controller(self, controller):
        self.controller = controller

    def reset_table(self):
        for item in self.tree.get_children():
            self.tree.delete(item)

    def fill_analysis_table(self, analysis_df):
        self.reset_table()
        # Insert new rows into the Treeview
        for row in analysis_df.itertuples(index=False):
            analysis_type = row[0]
            if len(analysis_type) > 1:
                analysis_type = self.analysis_type_abbreviation_dict[row[0]]
            self.tree.insert("", tk.END, values=(analysis_type, row[1], row[2], row[3]))

    def format_file_logic(self, logic):
        formatted_logic = []
        logic_dict = {"+": "and", "?": "or", "-": "not"}
        for idx, element in enumerate(logic):
            if len(element) > 1:
                if idx > 0:
                    logic_operator = logic_dict[element[0][0]]
                    formatted_logic.append(logic_operator)
                tag = element[1:]
                formatted_logic.append(tag)
            if element == "?":
                formatted_logic[idx - 1] = "or"
        return formatted_logic

    def format_imported_analysis_file(self, analysis_df):
        df_format = pd.DataFrame(
            columns=["analysis_type", "analysis_name", "group1_logic", "group2_logic"]
        )
        skip_row = False
        for row in range(analysis_df.shape[0]):
            if skip_row is True:
                skip_row = False
                continue

            analysis_type = analysis_df.iloc[row, 0]
            if analysis_type in ["B", "C"]:
                analysis_name = (
                    f"{analysis_df.iloc[row,1]} vs. {analysis_df.iloc[row+1,1]}"
                )
                group1_logic = list(
                    analysis_df.iloc[row, 2:].dropna().str.lower().str.strip()
                )
                group1_logic = self.format_file_logic(group1_logic)
                group2_logic = list(
                    analysis_df.iloc[row + 1, 2:].dropna().str.lower().str.strip()
                )
                group2_logic = self.format_file_logic(group2_logic)
                new_row = pd.DataFrame(
                    [
                        {
                            "analysis_type": analysis_type,
                            "analysis_name": analysis_name,
                            "group1_logic": group1_logic,
                            "group2_logic": group2_logic,
                        }
                    ]
                )
                skip_row = True
            else:
                analysis_name = analysis_df.iloc[row, 1]
                group1_logic = list(
                    analysis_df.iloc[row, 2:].dropna().str.lower().str.strip()
                )
                group1_logic = self.format_file_logic(group1_logic)
                new_row = pd.DataFrame(
                    [
                        {
                            "analysis_type": analysis_type,
                            "analysis_name": analysis_name,
                            "group1_logic": group1_logic,
                            "group2_logic": "----",
                        }
                    ]
                )

            df_format = pd.concat([df_format, new_row], ignore_index=True)
        return df_format



import logging
from pathlib import Path

import nibabel as nb
import numpy as np
from joblib import Parallel, delayed

from jale.core.utils.compute import (
    compute_ale,
    compute_permuted_ale_diff,
    compute_sig_diff,
)
from jale.core.utils.folder_setup import folder_setup
from jale.core.utils.plot_and_save import plot_and_save
from jale.core.utils.template import BRAIN_ARRAY_SHAPE

logger = logging.getLogger("ale_logger")


def contrast(
    project_path,
    meta_names,
    significance_threshold=0.0,
    null_repeats=10000,
    nprocesses=2,
):
    """
    Compute and save statistical contrasts and conjunctions for meta-analyses.

    This function calculates positive and negative contrasts, as well as conjunctions,
    between two meta-analyses specified by `meta_names`. If these results are already
    available, they are loaded from the saved files. Otherwise, the function computes
    the contrasts by estimating a null distribution through permutation testing, identifies
    significant voxels, and saves the results as NIfTI images.

    Parameters
    ----------
    project_path : str or Path
        Path to the project directory containing the "Results" folder.
    meta_names : list of str
        Names of the meta-analyses to compare; expects two names in the list.
    significance_threshold : float, optional
        Significance threshold for identifying significant voxels, by default 0..
    null_repeats : int, optional
        Number of permutations for generating the null distribution, by default 10000.
    nprocesses : int, optional
        Number of parallel processes for permutation testing, by default 4.

    Returns
    -------
    None
        The function performs computations and saves the results as NIfTI files in the
        specified `project_path` directory.
    """

    # set results folder as path
    folder_setup(project_path, "Contrast")
    project_path = (Path(project_path) / "Results/").resolve()

    ma1 = np.load(project_path / f"MainEffect/{meta_names[0]}_ma.npz")["arr_0"]
    ale1 = compute_ale(ma1)
    n_meta_group1 = ma1.shape[0]

    ma2 = np.load(project_path / f"MainEffect/{meta_names[1]}_ma.npz")["arr_0"]
    ale2 = compute_ale(ma2)
    n_meta_group2 = ma2.shape[0]

    # Check if contrast has already been calculated
    if Path(
        project_path / f"Contrast/Volumes/{meta_names[0]}_vs_{meta_names[1]}.nii"
    ).exists():
        logger.info(
            f"{meta_names[0]} x {meta_names[1]} - contrast already computed. Skipping."
        )  # noqa
    else:
        logger.info(f"{meta_names[0]} x {meta_names[1]} - Computing positive contrast.")  # noqa
        main_effect1 = nb.loadsave.load(
            project_path / f"MainEffect/Volumes/{meta_names[0]}_cFWE.nii"
        ).get_fdata()  # type: ignore
        significance_mask1 = main_effect1 > 0
        if significance_mask1.sum() > 0:
            stacked_masked_ma = np.vstack(
                (ma1[:, significance_mask1], ma2[:, significance_mask1])
            )

            ale_difference1 = ale1 - ale2
            # estimate null distribution of difference values if studies
            # would be randomly assigned to either meta analysis
            null_difference1 = Parallel(n_jobs=nprocesses, verbose=2)(
                delayed(compute_permuted_ale_diff)(stacked_masked_ma, n_meta_group1)
                for i in range(null_repeats)
            )
            z1, sig_idxs1 = compute_sig_diff(
                ale_difference1[significance_mask1],
                null_difference1,
                significance_threshold,
            )

        else:
            logger.warning(f"{meta_names[0]}: No significant indices!")
            z1, sig_idxs1 = [], []

        logger.info(f"{meta_names[1]} x {meta_names[0]} - Computing negative contrast.")
        main_effect2 = nb.loadsave.load(
            project_path / f"MainEffect/Volumes/{meta_names[1]}_cFWE.nii"
        ).get_fdata()  # type: ignore
        significance_mask2 = main_effect2 > 0
        if significance_mask2.sum() > 0:
            stacked_masked_ma = np.vstack(
                (ma1[:, significance_mask2], ma2[:, significance_mask2])
            )
            ale_difference2 = ale2 - ale1
            null_difference2 = Parallel(n_jobs=nprocesses, verbose=2)(
                delayed(compute_permuted_ale_diff)(stacked_masked_ma, n_meta_group2)
                for i in range(null_repeats)
            )
            z2, sig_idxs2 = compute_sig_diff(
                ale_difference2[significance_mask2],
                null_difference2,
                significance_threshold,
            )

        else:
            logger.warning(f"{meta_names[1]}: No significant indices!")
            z2, sig_idxs2 = np.array([]), []

        logger.info(f"{meta_names[0]} vs {meta_names[1]} - Inference and printing.")
        contrast_arr = np.zeros(BRAIN_ARRAY_SHAPE)
        flat_mask1 = np.where(significance_mask1)
        contrast_arr[
            flat_mask1[0][sig_idxs1], flat_mask1[1][sig_idxs1], flat_mask1[2][sig_idxs1]
        ] = z1

        flat_mask2 = np.where(significance_mask2)
        contrast_arr[
            flat_mask2[0][sig_idxs2], flat_mask2[1][sig_idxs2], flat_mask2[2][sig_idxs2]
        ] = -z2
        plot_and_save(
            project_path / "Contrast",
            f"{meta_names[0]}_vs_{meta_names[1]}",
            contrast_arr,
        )

    # Check if conjunction has already been calculated
    if Path(
        project_path / f"Contrast/Volumes/{meta_names[0]}_AND_{meta_names[1]}.nii"
    ).exists():
        logger.info(
            f"{meta_names[0]} & {meta_names[1]} - Conjunction already computed. Skipping."
        )
    else:
        logger.info(f"{meta_names[0]} & {meta_names[1]} - Computing conjunction.")
        conj_arr = np.minimum(main_effect1, main_effect2)
        if conj_arr is not None:
            plot_and_save(
                project_path / "Contrast",
                f"{meta_names[0]}_AND_{meta_names[1]}",
                conj_arr,
            )

    logger.info(f"{meta_names[0]} & {meta_names[1]} - done!")



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from scipy.cluster.hierarchy import fcluster, leaves_list, linkage
from scipy.spatial.distance import squareform
from scipy.stats import pearsonr, spearmanr
from sklearn.cluster import KMeans
from sklearn.metrics import (
    adjusted_rand_score,
    calinski_harabasz_score,
    silhouette_score,
)
from sklearn.metrics.cluster import entropy, mutual_info_score
from sklearn.utils import resample
from sklearn_extra.cluster import KMedoids

from jale.core.utils.compute import compute_ma
from jale.core.utils.folder_setup import folder_setup
from jale.core.utils.kernel import create_kernel_array
from jale.core.utils.template import GM_PRIOR


def clustering(
    project_path,
    exp_df,
    meta_name,
    correlation_type="spearman",  # spearman or pearson
    clustering_method="hierarchical",  # hierarchical or k-means
    linkage_method="complete",  # complete or average
    max_clusters=10,
    subsample_fraction=0.9,
    sampling_iterations=1000,
    null_iterations=1000,
):
    folder_setup(project_path, "MA_Clustering")
    kernels = create_kernel_array(exp_df)

    ma = compute_ma(exp_df.Coordinates.values, kernels)
    ma_gm_masked = ma[:, GM_PRIOR]

    if correlation_type == "spearman":
        correlation_matrix, _ = spearmanr(ma_gm_masked, axis=1)
    elif correlation_type == "pearson":
        correlation_matrix, _ = pearsonr(ma_gm_masked, axis=1)
    else:
        raise ValueError("Invalid correlation_type. Choose 'spearman' or 'pearson'.")

    plot_cor_matrix(project_path, correlation_matrix, linkage_method=linkage_method)

    (
        silhouette_scores,
        calinski_harabasz_scores,
        adjusted_rand_index,
        variation_of_information,
    ) = compute_clustering(
        meta_name,
        project_path,
        correlation_matrix,
        linkage_method=linkage_method,
        clustering_method=clustering_method,
        max_clusters=max_clusters,
        subsample_fraction=subsample_fraction,
        sampling_iterations=sampling_iterations,
    )

    null_silhouette_scores, null_calinski_harabasz_scores = compute_permute_clustering(
        meta_name,
        project_path,
        exp_df,
        kernels,
        max_clusters=max_clusters,
        null_iterations=null_iterations,
    )

    silhouette_z, alinski_harabasz_z = compute_metrics_z(
        silhouette_scores,
        calinski_harabasz_scores,
        null_silhouette_scores,
        null_calinski_harabasz_scores,
    )

    plot_clustering_metrics(
        project_path,
        silhouette_scores_z=silhouette_z,
        calinski_harabasz_scores_z=alinski_harabasz_z,
        adjusted_rand_index_z=adjusted_rand_index,
        voi_scores_z=variation_of_information,
    )

    save_clustering_metrics(
        project_path,
        silhouette_scores=silhouette_scores,
        silhouette_scores_z=silhouette_z,
        calinski_harabasz_scores=calinski_harabasz_scores,
        calinski_harabasz_scores_z=alinski_harabasz_z,
        adjusted_rand_index_z=adjusted_rand_index,
        voi_scores_z=variation_of_information,
    )


def compute_clustering(
    meta_name,
    project_path,
    correlation_matrix,
    clustering_method="hierarchical",
    linkage_method="complete",
    max_clusters=10,
    subsample_fraction=0.9,
    sampling_iterations=500,
):
    # Convert correlation matrix to correlation distance (1 - r)
    correlation_distance = 1 - correlation_matrix

    silhouette_scores = np.empty((max_clusters - 1, sampling_iterations))
    calinski_harabasz_scores = np.empty((max_clusters - 1, sampling_iterations))
    adjusted_rand_index = np.empty((max_clusters - 1, sampling_iterations))
    variation_of_information = np.empty((max_clusters - 1, sampling_iterations))

    # Iterate over different values of k, compute cluster metrics
    for k in range(2, max_clusters + 1):
        for i in range(sampling_iterations):
            # Resample indices for subsampling
            resampled_indices = resample(
                np.arange(correlation_matrix.shape[0]),
                replace=False,
                n_samples=int(subsample_fraction * correlation_matrix.shape[0]),
            )
            resampled_correlation = correlation_matrix[
                np.ix_(resampled_indices, resampled_indices)
            ]
            resampled_distance = correlation_distance[
                np.ix_(resampled_indices, resampled_indices)
            ]

            # Ensure diagonal is zero for distance matrix
            np.fill_diagonal(resampled_distance, 0)

            if clustering_method == "hierarchical":
                # Convert to condensed form for hierarchical clustering
                condensed_resampled_distance = squareform(
                    resampled_distance, checks=False
                )
                # Perform hierarchical clustering
                Z = linkage(condensed_resampled_distance, method=linkage_method)
                cluster_labels = fcluster(Z, k, criterion="maxclust")
            elif clustering_method == "kmeans":
                # Perform K-Means clustering
                kmeans = KMeans(n_clusters=k, random_state=i).fit(resampled_correlation)
                cluster_labels = kmeans.labels_
            else:
                raise ValueError(
                    "Invalid clustering_method. Choose 'hierarchical' or 'kmeans'."
                )

            # Silhouette Score
            silhouette_avg = silhouette_score(
                resampled_correlation
                if clustering_method == "kmeans"
                else resampled_distance,
                cluster_labels,
                metric="euclidean" if clustering_method == "kmeans" else "precomputed",
            )
            silhouette_scores[k - 2, i] = silhouette_avg

            # Calinski-Harabasz Index
            calinski_harabasz_avg = calinski_harabasz_score(
                resampled_correlation, cluster_labels
            )
            calinski_harabasz_scores[k - 2, i] = calinski_harabasz_avg

            # K-Medoids for comparison labels in adjusted rand and variation of information
            kmedoids = KMedoids(n_clusters=k, metric="precomputed").fit(
                resampled_distance
            )
            vof_labels = kmedoids.labels_

            # Adjusted Rand Score
            adjusted_rand_avg = adjusted_rand_score(cluster_labels, vof_labels)
            adjusted_rand_index[k - 2, i] = adjusted_rand_avg

            # Compute Variation of Information
            vi_score = compute_variation_of_information(cluster_labels, vof_labels)
            variation_of_information[k - 2, i] = vi_score

    # Save results
    np.save(
        project_path / f"Results/MA_Clustering/{meta_name}_silhouette_scores.npy",
        silhouette_scores,
    )
    np.save(
        project_path
        / f"Results/MA_Clustering/{meta_name}_calinski_harabasz_scores.npy",
        calinski_harabasz_scores,
    )
    np.save(
        project_path / f"Results/MA_Clustering/{meta_name}_adjusted_rand_index.npy",
        adjusted_rand_index,
    )
    np.save(
        project_path
        / f"Results/MA_Clustering/{meta_name}_variation_of_information.npy",
        variation_of_information,
    )

    return (
        silhouette_scores,
        calinski_harabasz_scores,
        adjusted_rand_index,
        variation_of_information,
    )


def compute_permute_clustering(
    meta_name, project_path, exp_df, kernels, max_clusters, null_iterations
):
    null_silhouette_scores = np.empty((max_clusters - 1, null_iterations))
    null_calinski_harabasz_scores = np.empty((max_clusters - 1, null_iterations))

    for n in range(null_iterations):
        coords_stacked = np.vstack(exp_df.Coordinates.values)
        shuffled_coords = []
        for exp in np.arange(exp_df.shape[0]):
            K = exp_df.loc[exp, "NumberOfFoci"]
            # Step 1: Randomly sample K unique row indices
            sample_indices = np.random.choice(
                coords_stacked.shape[0], size=K, replace=False
            )
            # Step 2: Extract the sampled rows using the sampled indices
            sampled_rows = coords_stacked[sample_indices]
            shuffled_coords.append(sampled_rows)
            # Step 3: Delete the sampled rows from the original array
            coords_stacked = np.delete(coords_stacked, sample_indices, axis=0)

        null_ma = compute_ma(shuffled_coords, kernels)
        ma_gm_masked = null_ma[:, GM_PRIOR]
        correlation_matrix, _ = spearmanr(ma_gm_masked, axis=1)
        correlation_matrix = np.nan_to_num(
            correlation_matrix, nan=0, posinf=0, neginf=0
        )
        correlation_distance = 1 - correlation_matrix
        condensed_distance = squareform(correlation_distance, checks=False)
        Z = linkage(condensed_distance, method="average")

        for k in range(2, max_clusters + 1):
            # Step 5: Extract clusters for k clusters
            cluster_labels = fcluster(Z, k, criterion="maxclust")

            # Silhouette Score
            silhouette_avg = silhouette_score(
                correlation_distance, cluster_labels, metric="precomputed"
            )
            null_silhouette_scores[k - 2, n] = silhouette_avg

            # Calinski-Harabasz Index
            calinski_harabasz_avg = calinski_harabasz_score(
                correlation_matrix, cluster_labels
            )
            null_calinski_harabasz_scores[k - 2, n] = calinski_harabasz_avg

    np.save(
        project_path / f"Results/MA_Clustering/{meta_name}_null_silhouette_scores.npy",
        null_silhouette_scores,
    )
    np.save(
        project_path
        / f"Results/MA_Clustering/{meta_name}_null_calinski_harabasz_scores.npy",
        null_calinski_harabasz_scores,
    )

    return (
        null_silhouette_scores,
        null_calinski_harabasz_scores,
    )


def compute_variation_of_information(labels_true, labels_pred):
    """
    Compute the Variation of Information (VI) metric.

    Parameters:
    labels_true (array-like): Ground truth cluster labels.
    labels_pred (array-like): Predicted cluster labels.

    Returns:
    float: VI score.
    """
    # Compute entropy for each clustering
    H_true = entropy(np.bincount(labels_true))
    H_pred = entropy(np.bincount(labels_pred))

    # Compute mutual information
    I_uv = mutual_info_score(labels_true, labels_pred)

    # Compute Variation of Information
    return H_true + H_pred - 2 * I_uv


def compute_metrics_z(
    silhouette_scores,
    calinski_harabasz_scores,
    null_silhouette_scores,
    null_calinski_harabasz_scores,
):
    silhouette_scores_avg = np.average(silhouette_scores, axis=1)
    calinski_harabasz_scores_avg = np.average(calinski_harabasz_scores, axis=1)

    null_silhouette_scores_avg = np.average(null_silhouette_scores, axis=1)
    null_calinski_harabasz_scores_avg = np.average(
        null_calinski_harabasz_scores, axis=1
    )

    silhouette_z = (silhouette_scores_avg - null_silhouette_scores_avg) / np.std(
        null_silhouette_scores
    )
    alinski_harabasz_z = (
        calinski_harabasz_scores_avg - null_calinski_harabasz_scores_avg
    ) / np.std(null_calinski_harabasz_scores)

    return silhouette_z, alinski_harabasz_z


def plot_cor_matrix(project_path, correlation_matrix, linkage_method="average"):
    # Perform hierarchical clustering
    linkage_matrix = linkage(correlation_matrix, method=linkage_method)

    # Get the ordering of rows/columns
    ordered_indices = leaves_list(linkage_matrix)

    # Reorder the correlation matrix
    sorted_correlation_matrix = correlation_matrix[ordered_indices][:, ordered_indices]
    plt.figure(figsize=(8, 6))
    sns.heatmap(sorted_correlation_matrix, cmap="RdBu_r", center=0, vmin=-1, vmax=1)

    # Add title and labels
    plt.title("Correlation Matrix with Custom Colormap")
    plt.xlabel("Experiments")
    plt.xticks(ticks=[])
    plt.ylabel("Experiments")
    plt.yticks(ticks=[])

    plt.savefig(project_path / "Results/MA_Clustering/correlation_matrix.png")


def plot_clustering_metrics(
    project_path,
    silhouette_scores_z,
    calinski_harabasz_scores_z,
    adjusted_rand_index,
    variation_of_info,
):
    plt.figure(figsize=(12, 8))

    # Plot Silhouette Scores
    plt.subplot(4, 1, 1)
    plt.plot(silhouette_scores_z, marker="o")
    plt.title("Silhouette Scores Z")
    plt.xlabel("Number of Clusters")
    plt.xticks(ticks=range(len(silhouette_scores_z)), labels=range(2, 11))
    plt.ylabel("Z-Score")
    plt.grid()

    # Plot Calinski-Harabasz Scores
    plt.subplot(4, 1, 2)
    plt.plot(calinski_harabasz_scores_z, marker="o")
    plt.title("Calinski-Harabasz Scores Z")
    plt.xlabel("Number of Clusters")
    plt.xticks(ticks=range(len(calinski_harabasz_scores_z)), labels=range(2, 11))
    plt.ylabel("Z-Score")
    plt.grid()

    # Plot Adjusted Rand Index
    plt.subplot(4, 1, 3)
    plt.plot(adjusted_rand_index, marker="o")
    plt.title("Adjusted Rand Index")
    plt.xlabel("Number of Clusters")
    plt.xticks(ticks=range(len(adjusted_rand_index)), labels=range(2, 11))
    plt.ylabel("aRI-Score")
    plt.grid()

    # Plot Variation of Information
    plt.subplot(4, 1, 4)
    plt.plot(variation_of_info, marker="o")
    plt.title("Variation of Information")
    plt.xlabel("Number of Clusters")
    plt.xticks(ticks=range(len(variation_of_info)), labels=range(2, 11))
    plt.ylabel("VI-Score")
    plt.grid()

    plt.tight_layout()
    plt.savefig(project_path / "Results/MA_Clustering/clustering_metrics.png")
    plt.show()


def save_clustering_metrics(
    project_path,
    silhouette_scores,
    silhouette_scores_z,
    calinski_harabasz_scores,
    calinski_harabasz_scores_z,
    adjusted_rand_index,
    variation_of_info,
):
    metrics_df = pd.DataFrame(
        {
            "Number of Clusters": range(2, 11),
            "Silhouette Scores": silhouette_scores,
            "Silhouette Scores Z": silhouette_scores_z,
            "Calinski-Harabasz Scores": calinski_harabasz_scores,
            "Calinski-Harabasz Scores Z": calinski_harabasz_scores_z,
            "Adjusted Rand Index": adjusted_rand_index,
            "Variation of Information": variation_of_info,
        }
    )
    metrics_df.to_csv(
        project_path / "Results/MA_Clustering/clustering_metrics.csv", index=False
    )



import customtkinter


class Sidebar_Frame(customtkinter.CTkFrame):
    def __init__(self, master, corner_radius: int = 0):
        super().__init__(master, corner_radius=corner_radius)
        self.add_analysis_window = None
        self.parameter_window = None
        self.parameter_warning_window = None

        self.select_config_file_button = customtkinter.CTkButton(
            master=self,
            text="Select ALE Config",
            command=self.select_config_file_button_event,
        )
        self.select_config_file_button.grid(row=0, column=0, padx=20, pady=(20, 20))

        self.run_analysis_button = customtkinter.CTkButton(
            master=self,
            text="Run Analysis",
            fg_color="green4",
            hover_color="dark green",
            command=self.run_analysis_button_event,
        )
        self.run_analysis_button.grid(row=6, column=0, padx=20, pady=10)

        self.stop_analysis_button = customtkinter.CTkButton(
            master=self,
            text="Stop Analysis",
            fg_color="red3",
            hover_color="red4",
            command=self.stop_analysis_button_event,
        )
        self.stop_analysis_button.grid(row=7, column=0, padx=20, pady=10)

        self.appearance_mode_label = customtkinter.CTkLabel(
            master=self, text="Appearance Mode:", anchor="w"
        )
        self.appearance_mode_label.grid(row=8, column=0, padx=20, pady=(10, 0))
        self.appearance_mode_optionemenu = customtkinter.CTkOptionMenu(
            master=self,
            values=["Light", "Dark", "System"],
            command=self.change_appearance_mode_event,
        )
        self.appearance_mode_optionemenu.set("Dark")
        self.appearance_mode_optionemenu.grid(row=9, column=0, padx=20, pady=(10, 10))

        self.scaling_label = customtkinter.CTkLabel(
            master=self, text="UI Scaling:", anchor="w"
        )
        self.scaling_label.grid(row=10, column=0, padx=20, pady=(10, 0))
        self.scaling_optionemenu = customtkinter.CTkOptionMenu(
            master=self,
            values=["80%", "90%", "100%", "110%", "120%"],
            command=self.change_scaling_event,
        )
        self.scaling_optionemenu.set("100%")
        self.scaling_optionemenu.grid(row=11, column=0, padx=20, pady=(10, 20))

    def set_controller(self, controller):
        self.controller = controller

    def select_config_file_button_event(self):
        self.controller.config_to_gui()
        return

    def run_analysis_button_event(self):
        self.controller.run_analysis()
        return

    def stop_analysis_button_event(self):
        self.controller.stop_analysis()
        return

    def change_appearance_mode_event(self, new_appearance_mode: str):
        customtkinter.set_appearance_mode(new_appearance_mode)

    def change_scaling_event(self, new_scaling: str):
        new_scaling_float = int(new_scaling.replace("%", "")) / 100
        customtkinter.set_widget_scaling(new_scaling_float)



"""This module provides a function to set up a directory structure for storing analysis results."""

from pathlib import Path


def folder_setup(path, type_of_analysis):
    """
    Set up a directory structure for storing analysis results.

    This function creates a nested folder structure under the specified path
    for organizing result files related to various analyses. The structure
    depends on the specified type of analysis.

    Parameters
    ----------
    path : Path or str
        Base path where the directory structure should be created.

    type_of_analysis : str
        Type of analysis to set up folders for. Options are:
        - "MainEffect"
        - "Probabilistic"
        - "Contrast"
        - "BalancedContrast"
        - "MA_Clustering"
        - "ROI"

    Returns
    -------
    None
    """

    # Convert path to Path object if it's a string
    path = Path(path)

    # Define folder structures for each analysis type
    analysis_folders = {
        "MainEffect": [
            "Results/MainEffect/Volumes",
            "Results/MainEffect/Contribution",
            "Results/MainEffect/NullDistributions",
            "Results/MainEffect/Figures",
        ],
        "Probabilistic": [
            "Results/Probabilistic/Volumes",
            "Results/Probabilistic/NullDistributions",
            "Results/Probabilistic/Figures",
        ],
        "Contrast": [
            "Results/Contrast/Volumes",
            "Results/Contrast/NullDistributions",
            "Results/Contrast/Figures",
        ],
        "BalancedContrast": [
            "Results/BalancedContrast/Volumes",
            "Results/BalancedContrast/NullDistributions",
            "Results/BalancedContrast/Figures",
        ],
        "MA_Clustering": [
            "Results/MA_Clustering",
        ],
        "ROI": [
            "Results/MainEffect/ROI",
        ],
    }

    # Validate the type_of_analysis
    if type_of_analysis not in analysis_folders:
        raise ValueError(
            f"Invalid analysis type: {type_of_analysis}. "
            "Valid options are: MainEffect, Probabilistic, Contrast, BalancedContrast, MA_Clustering, ROI."
        )

    # Get the folder structure for the selected analysis type
    folders_to_create = analysis_folders[type_of_analysis]

    # Create the directories for the selected analysis type
    for folder in folders_to_create:
        (path / folder).mkdir(parents=True, exist_ok=True)



import logging
import pickle
from pathlib import Path

import nibabel as nb
import numpy as np
import pandas as pd
from joblib import Parallel, delayed

from jale.core.utils.compute import (
    compute_ma,
    compute_monte_carlo_null,
    compute_sub_ale,
    generate_unique_subsamples,
)
from jale.core.utils.folder_setup import folder_setup
from jale.core.utils.kernel import create_kernel_array
from jale.core.utils.plot_and_save import plot_and_save

logger = logging.getLogger("ale_logger")


def probabilistic_ale(
    project_path,
    exp_df,
    meta_name,
    bin_steps=0.0001,
    cluster_forming_threshold=0.001,
    monte_carlo_iterations=5000,
    target_n=None,
    sample_n=2500,
    nprocesses=2,
):
    """
    Compute and save probabilistic ALE maps for a given meta-analysis.

    Parameters
    ----------
    project_path : str or Path
        Path to the project directory containing the "Results" folder.
    exp_df : pandas.DataFrame
        DataFrame containing experiment data, including coordinates and number of foci.
    meta_name : str
        Name of the meta-analysis, used for naming saved files.
    tfce_enabled : bool, optional
        Whether to compute TFCE-corrected maps, by default True.
    cutoff_predict_enabled : bool, optional
        If True, predicts statistical thresholds using ML models, by default True.
    bin_steps : float, optional
        Step size for defining histogram bins, by default 0.0001.
    cluster_forming_threshold : float, optional
        Threshold for forming clusters in ALE, by default 0.001.
    monte_carlo_iterations : int, optional
        Number of Monte Carlo iterations for null distribution simulation, by default 5000.
    target_n : int, optional
        Target number of subsamples for probabilistic ALE, by default None (uses full sample).
    sample_n : int, optional
        Number of subsamples to generate if `target_n` is specified, by default 2500.
    nprocesses : int, optional
        Number of parallel processes for computations, by default 2.

    Returns
    -------
    None
        The function performs computations and saves the results as NIfTI files in the
        specified `project_path` directory.
    """
    logger.info(
        f"{meta_name} : {exp_df.shape[0]} experiments; average of {exp_df.Subjects.mean():.2f} subjects per experiment"
    )

    folder_setup(project_path, "Probabilistic")
    # set cv results folder as path
    project_path = (Path(project_path) / "Results/Probabilistic/").resolve()

    # calculate smoothing kernels for each experiment
    kernels = create_kernel_array(exp_df)
    np.save(project_path / f"{meta_name}_kernels", kernels)

    # calculate maximum possible ale value to set boundaries for histogram bins
    max_ma = np.prod([1 - np.max(kernel) for kernel in kernels])

    # define bins for histogram
    bin_edges = np.arange(0.00005, 1 - max_ma + 0.001, bin_steps)
    bin_centers = np.arange(0, 1 - max_ma + 0.001, bin_steps)
    step = int(1 / bin_steps)

    # Save included experiments for provenance tracking
    print_df = pd.DataFrame(
        {
            "Experiment": exp_df.Articles.values,
            "Number of Foci": exp_df.NumberOfFoci.values,
        }
    )
    print_df.to_csv(
        project_path / f"{meta_name}_included_experiments.csv", index=False, sep="\t"
    )

    # MA calculation
    ma = compute_ma(exp_df.Coordinates.values, kernels)
    np.savez_compressed(project_path / f"{meta_name}_ma", ma)

    # subsampling or probabilistic ALE
    logger.info(f"{meta_name} - entering probabilistic ALE routine.")
    # Check whether monte-carlo cutoff has been calculated before
    if Path(
        project_path / f"NullDistributions/{meta_name}_montecarlo_{target_n}.pickle"
    ).exists():
        logger.info(f"{meta_name} - loading cv cluster cut-off.")
        with open(
            project_path
            / f"NullDistributions/{meta_name}_montecarlo_{target_n}.pickle",
            "rb",
        ) as f:
            cfwe_null = pickle.load(f)
            subsampling_cfwe_threshold = np.percentile(cfwe_null, 95)
    else:
        logger.info(f"{meta_name} - computing cv cluster cut-off.")
        _, cfwe_null, _ = zip(
            *Parallel(n_jobs=nprocesses, verbose=2)(
                delayed(compute_monte_carlo_null)(
                    num_foci=exp_df.NumberOfFoci,
                    kernels=kernels,
                    bin_edges=bin_edges,
                    bin_centers=bin_centers,
                    step=step,
                    cluster_forming_threshold=cluster_forming_threshold,
                    target_n=target_n,
                    tfce_enabled=False,
                )
                for i in range(monte_carlo_iterations)
            )
        )

        subsampling_cfwe_threshold = np.percentile(cfwe_null, 95)
        with open(
            project_path
            / f"NullDistributions/{meta_name}_montecarlo_{target_n}.pickle",
            "wb",
        ) as f:
            pickle.dump(cfwe_null, f)
    if Path(project_path / f"Volumes/{meta_name}_sub_ale_{target_n}.nii").exists():
        logger.info(f"{meta_name} - loading cv ale")

        ale_mean = nb.load(
            project_path / f"Volumes/{meta_name}_sub_ale_{target_n}.nii"
        ).get_fdata()

    else:
        logger.info(f"{meta_name} - computing cv ale.")

        samples = generate_unique_subsamples(
            total_n=exp_df.shape[0], target_n=target_n, sample_n=sample_n
        )
        ale_mean = compute_sub_ale(
            samples,
            ma,
            subsampling_cfwe_threshold,
            bin_edges,
            bin_centers,
            step,
            cluster_forming_threshold,
        )
        plot_and_save(project_path, f"{meta_name}_sub_ale_{target_n}", ale_mean)

        logger.info(f"{meta_name} - probabilistic ALE done!")



from setuptools import find_packages, setup

setup(
    name="jale",  # Unique name for your package
    version="0.1.30",  # Version number
    description="Package allowing users to run Activation Likelihood Estimation Meta-Analysis",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Lennart Frahm",
    author_email="l.frahm@mailbox.org",
    url="https://github.com/LenFrahm/JALE",
    license="MIT",
    packages=find_packages(),  # Automatically find packages in the project
    install_requires=[  # List of dependencies
        "customtkinter==5.2.2",
        "joblib>=1.3.2",
        "nibabel>=5.2.1",
        "numpy==1.26.4",
        "pandas>=2.0.3",
        "PyYAML>=6.0.2",
        "scipy>=1.10.1",
        "xgboost>=2.1.2",
        "openpyxl>=3.1.5",
        "scikit-learn-extra>=0.3.0",
        "matplotlib>=3.7.1",
        "seaborn>=0.12.2",
        "nilearn>=0.10.4",
    ],
    include_package_data=True,
    classifiers=[  # Classifiers help users find your project
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.9, <3.13",
)



import nibabel as nb
import numpy as np
from nilearn import plotting

from jale.core.utils.template import GM_PRIOR, MNI_AFFINE


def plot_and_save(project_path, analysis_name, arr):
    """
    Save a brain data array as a NIfTI file.

    Applies a mask to the array based on a prior and saves it as a NIfTI image.

    Parameters
    ----------
    arr : numpy.ndarray
        Brain data array to save.
    nii_path : str or Path
        Path to save the NIfTI file.

    Returns
    -------
    None
    """

    # Function that takes brain array and transforms it to NIFTI1 format
    # Saves it as a Nifti file
    arr = np.nan_to_num(arr, nan=0, posinf=0, neginf=0)
    arr[GM_PRIOR == 0] = 0
    nii_img = nb.nifti1.Nifti1Image(arr, MNI_AFFINE)
    if arr.any() > 0:
        plotting.plot_stat_map(
            nii_img, output_file=project_path / f"Figures/{analysis_name}"
        )
        nb.loadsave.save(nii_img, project_path / f"Volumes/{analysis_name}")
    else:
        nb.loadsave.save(nii_img, project_path / f"Volumes/{analysis_name}_empty")



import logging
from pathlib import Path

import numpy as np

from jale.core.analyses.balanced_contrast import balanced_contrast
from jale.core.analyses.clustering import clustering
from jale.core.analyses.contrast import contrast
from jale.core.analyses.main_effect import main_effect
from jale.core.analyses.probabilistic import probabilistic_ale
from jale.core.analyses.roi import roi_ale
from jale.core.utils.compile_experiments import compile_experiments
from jale.core.utils.contribution import contribution
from jale.core.utils.input import load_config, load_dataframes
from jale.core.utils.logger import setup_logger


def run_main_effect(analysis_df, row_idx, project_path, params, exp_all_df, tasks):
    """
    Run a main-effect analysis based on the analysis dataframe and experiment info.

    Parameters
    ----------
    analysis_df : pandas.DataFrame
        DataFrame containing the analysis information, including meta-analysis names
        and conditions for experiment selection.
    row_idx : int
        Index of the current row in the analysis dataframe.
    project_path : str or Path
        Path to the project directory where results are saved.
    params : dict
        Dictionary of parameters for analysis, including Monte Carlo iterations and
        subsample size.
    exp_all_df : pandas.DataFrame
        DataFrame containing all available experimental data.
    tasks : pandas.DataFrame
        DataFrame containing task information used for compiling experiments.

    Returns
    -------
    None
        The function performs computations and saves the results.
    """
    logger = logging.getLogger("ale_logger")
    meta_name = analysis_df.iloc[row_idx, 1]

    result_path = project_path / f"Results/MainEffect/Volumes/{meta_name}_cFWE.nii"
    if result_path.exists():
        logger.info(f"Main Effect results for {meta_name} already exist.")
        return

    logger.info("Running Main-Effect Analysis")
    conditions = analysis_df.iloc[row_idx, 2:].dropna().to_list()
    exp_idxs, masks, mask_names = compile_experiments(conditions, tasks)
    exp_df = exp_all_df.loc[exp_idxs].reset_index(drop=True)

    main_effect(
        project_path,
        exp_df,
        meta_name,
        tfce_enabled=params["tfce_enabled"],
        cutoff_predict_enabled=params["cutoff_predict_enabled"],
        bin_steps=params["bin_steps"],
        cluster_forming_threshold=params["cluster_forming_threshold"],
        monte_carlo_iterations=params["monte_carlo_iterations"],
        nprocesses=params["nprocesses"],
    )
    contribution(
        project_path, exp_df, exp_idxs, meta_name, tasks, params["tfce_enabled"]
    )

    if masks:
        for idx, mask in enumerate(masks):
            roi_ale(
                project_path,
                exp_df,
                meta_name,
                mask,
                mask_names[idx],
                monte_carlo_iterations=params["monte_carlo_iterations"],
            )


def run_probabilistic_ale(
    analysis_df, row_idx, project_path, params, exp_all_df, tasks
):
    """
    Run a probabilistic Activation Likelihood Estimation (ALE) analysis.

    Parameters
    ----------
    analysis_df : pandas.DataFrame
        DataFrame containing the analysis information, including meta-analysis
        names and conditions for experiment selection.
    row_idx : int
        Index of the current row in the analysis dataframe.
    project_path : str or Path
        Path to the project directory where results are saved.
    params : dict
        Dictionary of parameters for analysis, including Monte Carlo iterations
        and subsample size.
    exp_all_df : pandas.DataFrame
        DataFrame containing all available experimental data.
    tasks : pandas.DataFrame
        DataFrame containing task information used for compiling experiments.

    Returns
    -------
    None
        The function performs computations and saves the results.
    """
    logger = logging.getLogger("ale_logger")
    meta_name = analysis_df.iloc[row_idx, 1]

    target_n = (
        int(analysis_df.iloc[row_idx, 0][1:])
        if len(analysis_df.iloc[row_idx, 0]) > 1
        else None
    )

    result_path = (
        project_path
        / f"Results/Probabilistic/Volumes/{meta_name}_sub_ale_{target_n}.nii"
    )
    if result_path.exists():
        logger.info(f"Probabilistic ALE results for {meta_name} already exist.")
        return

    logger.info("Running Probabilistic ALE")
    conditions = analysis_df.iloc[row_idx, 2:].dropna().to_list()
    exp_idxs, _, _ = compile_experiments(conditions, tasks)
    exp_df = exp_all_df.loc[exp_idxs].reset_index(drop=True)

    if target_n:
        probabilistic_ale(
            project_path,
            exp_df,
            meta_name,
            target_n=target_n,
            monte_carlo_iterations=params["monte_carlo_iterations"],
            sample_n=params["subsample_n"],
            nprocesses=params["nprocesses"],
        )
    else:
        logger.warning(f"{meta_name}: Need to specify subsampling N")


def run_contrast_analysis(
    analysis_df, row_idx, project_path, params, exp_all_df, tasks
):
    """
    Run a contrast analysis between two meta-analyses.

    Parameters
    ----------
    analysis_df : pandas.DataFrame
        DataFrame containing the analysis information.
    row_idx : int
        Index of the current row in the DataFrame.
    project_path : str or Path
        Path to the project directory.
    params : dict
        Dictionary of parameters for analysis, including significance threshold and
        number of permutations.
    exp_all_df : pandas.DataFrame
        DataFrame containing all experiment data.
    tasks : pandas.DataFrame
        DataFrame containing task information.

    Returns
    -------
    None
        The function performs computations and saves the results as NIfTI files in the
        specified `project_path` directory.
    """
    meta_names, exp_dfs, exp_idxs = setup_contrast_data(
        analysis_df, row_idx, exp_all_df, tasks
    )

    for idx, meta_name in enumerate(meta_names):
        result_path = project_path / f"Results/MainEffect/Volumes/{meta_name}_cFWE.nii"
        if not result_path.exists():
            logger = logging.getLogger("ale_logger")
            logger.info(
                f"Running main effect for {meta_name} as prerequisite for contrast analysis"
            )
            main_effect(
                project_path,
                exp_dfs[idx],
                meta_name,
                tfce_enabled=params["tfce_enabled"],
                cutoff_predict_enabled=params["cutoff_predict_enabled"],
                bin_steps=params["bin_steps"],
                cluster_forming_threshold=params["cluster_forming_threshold"],
                monte_carlo_iterations=params["monte_carlo_iterations"],
                nprocesses=params["nprocesses"],
            )
            contribution(
                project_path,
                exp_dfs[idx],
                exp_idxs[idx],
                meta_name,
                tasks,
                params["tfce_enabled"],
            )

    exp_overlap = set(exp_dfs[0].index) & set(exp_dfs[1].index)
    exp_dfs = [exp_dfs[0].drop(exp_overlap), exp_dfs[1].drop(exp_overlap)]

    contrast(
        project_path,
        meta_names,
        significance_threshold=params["significance_threshold"],
        null_repeats=params["contrast_permutations"],
        nprocesses=params["nprocesses"],
    )


def run_balanced_contrast(
    analysis_df, row_idx, project_path, params, exp_all_df, tasks
):
    """
    Run a balanced contrast analysis using the provided experiment data.

    Parameters
    ----------
    analysis_df : pandas.DataFrame
        DataFrame containing the analysis information.
    row_idx : int
        Index of the current row in the DataFrame.
    project_path : str or Path
        Path to the project directory.
    params : dict
        Dictionary of parameters for analysis, including TFCE and cutoff prediction settings.
    exp_all_df : pandas.DataFrame
        DataFrame containing all experiment data.
    tasks : pandas.DataFrame
        DataFrame containing task information.

    Returns
    -------
    None
        The function performs computations and saves the results as NIfTI files in the
        specified `project_path` directory.
    """
    meta_names, exp_dfs, exp_idxs = setup_contrast_data(
        analysis_df, row_idx, exp_all_df, tasks
    )
    target_n = determine_target_n(analysis_df.iloc[row_idx, 0], exp_dfs)

    # Check if subsampling ALE were already run; if not - run them
    for idx, meta_name in enumerate(meta_names):
        result_path = (
            project_path
            / f"Results/MainEffect/Volumes/{meta_name}_sub_ale_{target_n}.nii"
        )
        if not result_path.exists():
            logger = logging.getLogger("ale_logger")
            logger.info(
                f"Running subsampling ale for {meta_name} as prerequisite for balanced contrast analysis"
            )
            probabilistic_ale(
                project_path,
                exp_dfs[idx],
                meta_name,
                target_n=target_n,
                monte_carlo_iterations=params["monte_carlo_iterations"],
                sample_n=params["subsample_n"],
                nprocesses=params["nprocesses"],
            )

    balanced_contrast(
        project_path,
        exp_dfs,
        meta_names,
        target_n,
        difference_iterations=params["difference_iterations"],
        monte_carlo_iterations=params["monte_carlo_iterations"],
        nprocesses=2,
    )


def setup_contrast_data(analysis_df, row_idx, exp_all_df, tasks):
    """
    Prepare experiment data for contrast analysis.

    Parameters
    ----------
    analysis_df : pandas.DataFrame
        DataFrame containing analysis information, including meta-analysis names
        and conditions for experiment selection.
    row_idx : int
        Index of the current row in the analysis dataframe from which to start
        extracting meta-analysis data.
    exp_all_df : pandas.DataFrame
        DataFrame containing all available experimental data.
    tasks : pandas.DataFrame
        DataFrame containing task information used for compiling experiments.

    Returns
    -------
    tuple
        A tuple containing:
        - list of str: Names of the meta-analyses for the selected rows.
        - list of pandas.DataFrame: DataFrames for the experiments corresponding
          to each meta-analysis.
    """
    meta_names = [analysis_df.iloc[row_idx, 1], analysis_df.iloc[row_idx + 1, 1]]
    conditions = [
        analysis_df.iloc[row_idx, 2:].dropna().to_list(),
        analysis_df.iloc[row_idx + 1, 2:].dropna().to_list(),
    ]
    exp_idxs1, _, _ = compile_experiments(conditions[0], tasks)
    exp_idxs2, _, _ = compile_experiments(conditions[1], tasks)

    exp_idxs = [exp_idxs1, exp_idxs2]

    exp_dfs = [
        exp_all_df.loc[exp_idxs1].reset_index(drop=True),
        exp_all_df.loc[exp_idxs2].reset_index(drop=True),
    ]
    return meta_names, exp_dfs, exp_idxs


def determine_target_n(row_value, exp_dfs):
    """
    Determine the target number of subsamples for analysis.

    Parameters
    ----------
    row_value : str
        A string value from the analysis dataframe indicating the target subsample size.
    exp_dfs : list of pandas.DataFrame
        List of DataFrames containing experiment data for different meta-analyses.

    Returns
    -------
    int
        The calculated target number of subsamples.
    """
    if len(row_value) > 1:
        return int(row_value[1:])
    n = [len(exp_dfs[0]), len(exp_dfs[1])]
    return int(min(np.floor(np.mean((np.min(n), 17))), np.min(n) - 2))


def run_ma_clustering(analysis_df, row_idx, project_path, params, exp_all_df, tasks):
    logger = logging.getLogger("ale_logger")
    logger.info("Running MA Clustering")

    meta_name = analysis_df.iloc[row_idx, 1]
    conditions = analysis_df.iloc[row_idx, 2:].dropna().to_list()
    exp_idxs, masks, mask_names = compile_experiments(conditions, tasks)
    exp_df = exp_all_df.loc[exp_idxs].reset_index(drop=True)

    logger.info(
        f"{meta_name} : {len(exp_idxs)} experiments; average of {exp_df.Subjects.mean():.2f} subjects per experiment"
    )

    clustering(
        project_path,
        exp_df,
        meta_name,
        max_clusters=params["max_clusters"],
        subsample_fraction=params["subsample_fraction"],
        sampling_iterations=params["sampling_iterations"],
        null_iterations=params["null_iterations"],
    )


def run_ale(yaml_path=None):
    # Load config and set up paths
    config = load_config(yaml_path)
    project_path = Path(yaml_path).parent
    # Create a logs folder (common across all analysis types)
    (project_path / "logs").mkdir(parents=True, exist_ok=True)
    # Initialize the logger
    logger = setup_logger(project_path)
    logger.info("Logger initialized and project setup complete.")

    params = config.get("parameters", {})
    clustering_params = config.get("clustering_parameters", {})
    exp_all_df, tasks, analysis_df = load_dataframes(project_path, config)

    # Main loop to process each row in the analysis dataframe
    for row_idx in range(analysis_df.shape[0]):
        # skip empty rows - indicate 2nd effect for contrast analysis
        if not isinstance(analysis_df.iloc[row_idx, 0], str):
            continue

        if analysis_df.iloc[row_idx, 0] == "M":
            run_main_effect(
                analysis_df, row_idx, project_path, params, exp_all_df, tasks
            )
        elif analysis_df.iloc[row_idx, 0][0] == "P":
            run_probabilistic_ale(
                analysis_df, row_idx, project_path, params, exp_all_df, tasks
            )
        elif analysis_df.iloc[row_idx, 0] == "C":
            run_contrast_analysis(
                analysis_df, row_idx, project_path, params, exp_all_df, tasks
            )
        elif analysis_df.iloc[row_idx, 0][0] == "B":
            run_balanced_contrast(
                analysis_df, row_idx, project_path, params, exp_all_df, tasks
            )
        elif analysis_df.iloc[row_idx, 0] == "Cluster":
            run_ma_clustering(
                analysis_df, row_idx, project_path, clustering_params, exp_all_df, tasks
            )

    logger.info("Analysis completed.")



import logging
from datetime import datetime
from pathlib import Path


def setup_logger(project_path: Path):
    """Initialize logging with a file handler in the project directory."""
    logger = logging.getLogger("ale_logger")

    # Prevent adding handlers multiple times
    if not logger.hasHandlers():
        logger.setLevel(logging.DEBUG)

        # Console handler
        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.INFO)
        console_format = logging.Formatter("%(levelname)s - %(message)s")
        console_handler.setFormatter(console_format)

        # File handler in the project directory
        start_time = datetime.now().strftime("%Y%m%d_%H%M")
        file_handler = logging.FileHandler(project_path / f"logs/{start_time}.log")
        file_handler.setLevel(logging.DEBUG)
        file_format = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        file_handler.setFormatter(file_format)

        # Add handlers to the logger
        logger.addHandler(console_handler)
        logger.addHandler(file_handler)

    return logger
